/**
 *  A3
 *
 *  By Vincent Petry - PVince81 at yahoo dot fr
 *
 *  ---------------------------------------------------------------------------
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.*
 */

(function() {
    var renderer = null,
        sprites = {},
        cornerSprites = {},
        timer = null,
        animations = [];

    function initSprites(callback) {
        var tex = renderer.createTexture("img/a3_sprites.png", function() {
            sprites = {
                block: renderer.createSprite(tex, 0, 0, 50, 50),
                marble: renderer.createSprite(tex, 50, 0, 50, 50),
                circle: renderer.createSprite(tex, 0, 50, 50, 50),
                rectangle: renderer.createSprite(tex, 50, 50, 50, 50),
                triangle: renderer.createSprite(tex, 100, 50, 50, 50),
                star: renderer.createSprite(tex, 150, 50, 50, 50),
                heart: renderer.createSprite(tex, 200, 50, 50, 50),
                cursor: renderer.createSprite(tex, 0, 100, 50, 50),
                chains: renderer.createSprite(tex, 50, 100, 50, 50)
            };
            if ( callback ) {
                callback();
            }
        });
        var tex2 = renderer.createTexture("img/a3_corners.png", function() {
            cornerSprites = {
                topleft: renderer.createSprite(tex2, 0, 0, 25, 25),
                top: renderer.createSprite(tex2, 25, 0, 25, 25),
                topright: renderer.createSprite(tex2, 100, 0, 25, 25),
                left: renderer.createSprite(tex2, 0, 25, 25, 25),
                right: renderer.createSprite(tex2, 100, 25, 25, 25),
                bottomleft: renderer.createSprite(tex2, 0, 100, 25, 25),
                bottom: renderer.createSprite(tex2, 25, 100, 25, 25),
                bottomright: renderer.createSprite(tex2, 100, 100, 25, 25),
                outertopleft: renderer.createSprite(tex2, 0, 0, 15, 15),
                outertopright: renderer.createSprite(tex2, 110, 0, 15, 15),
                outerbottomright: renderer.createSprite(tex2, 110, 110, 15, 15),
                outerbottomleft: renderer.createSprite(tex2, 0, 110, 15, 15),
                cross: renderer.createSprite(tex2, 150, 0, 25, 25),
                cross2: renderer.createSprite(tex2, 150, 25, 25, 25),
            }

            for ( var spriteName in cornerSprites ) {
                var sprite = cornerSprites[spriteName];
                sprite.marginX = -10;
                sprite.marginY = -10;
            }

            cornerSprites.outertopleft.marginX = 0;
            cornerSprites.outertopleft.marginY = 0;

            cornerSprites.outertopright.marginY = 0;
            cornerSprites.outerbottomleft.marginX = 0;
        });
    };

    var Constants = a3.Constants;

    var AnimationManager = function() {
        this.lastTime = new Date().getTime();
        this.animationGCTime = 2000;
        this._animations = [];
        this._handlers = {};
    };

    AnimationManager.prototype = {
        _animations: null,
        // map of objects to properties to animations
        _handlers: null,

        _addHandler: function( animation ) {
            var handler = this._handlers[animation.object.id];
            if ( !handler ) {
                handler = this._handlers[animation.object.id] = {};
            }

            var curAnim = handler[animation.prop];
            if ( curAnim && !curAnim.finished ) {
                // stop conflicting animation
                curAnim.stop();
            }
            handler[animation.prop] = animation;
        },

        _removeHandler: function(animation) {
            var handler = this._handlers[animation.object.id];
            if ( !handler ) {
                return;
            }

            if ( animation.object.removed ) {
                delete this._handlers[animation.object.id];
            }

            delete handler[animation.prop];
        },

        add: function( animation ) {
            if ( animation.finished ) {
                return;
            }
            this._addHandler(animation);
            this._animations.push(animation);
        },

        process: function() {
            var timeNow = new Date().getTime(),
                elapsed = timeNow - this.lastTime,
                animations = this._animations,
                dirty = false;

            for ( var i = 0; i < animations.length; i++ ) {
                if ( animations[i].step(elapsed) ) {
                    dirty = true;
                }
            }
            this.animationGCTime -= elapsed;
            if ( this.animationGCTime <= 0 ) {
                var newAnims = [];
                for ( var i = 0; i < animations.length; i++ ) {
                    var anim = animations[i];
                    if ( !anim.finished ) {
                        newAnims.push(animations[i]);
                    }
                    else {
                        this._removeHandler(anim);
                    }
                }
                animations = newAnims;
                this.animationGCTime = 2000;
            }
            this.lastTime = timeNow;
            return dirty;
        }
    }

    var animationManager = new AnimationManager();
    
    var NumberAnimation = function(object, prop, value, duration, callback) {
        this.id = makeId();
        this.startValue = object[prop];
        this.targetValue = value;
        this.progress = 0.0;
        this.duration = duration;
        this.object = object;
        this.prop = prop;
        this.finished = false;
        this.callback = callback;
        this.func = function(progress) {
            return progress;
        }
        if ( this.stepValue == 0 ) {
            this.finished = true;
            if ( callback ) {
                callback(this);
            }
            return;
        }
    };

    NumberAnimation.prototype = {
        stop: function() {
            this.finished = true;
        },

        step: function(delta) {
            if ( this.finished ) {
                return false;
            }
            this.progress += delta;
            if ( this.progress >= this.duration ) {
                this.progress = this.duration;
                this.finished = true;
            }
            this.object[this.prop] = this.startValue + (this.targetValue - this.startValue) * this.func(this.progress / this.duration);
            return true;
        }
    };

    var Element = function(sprite, x, y) {
        this.id = makeId();
        this.sprite = sprite;
        this.x = x || 0;
        this.y = y || 0;
        this.visible = true;
        this.alpha = 1.0;
        this.scale = 1.0;
    }

    Element.prototype = {
        render: function(opts) {
            if ( !this.sprite || !this.visible ) {
                return;
            }
            var alpha = this.alpha;
            if ( opts && opts.alpha !== undefined ) {
                alpha *= opts.alpha;
            }
            renderer.drawSprite(this.sprite, this.x, this.y, alpha, this.scale);
        },

        moveTo: function(targetX, targetY, duration) {
            if ( duration == 0 ) {
                this.x = targetX;
                this.y = targetY;
                return;
            }

            if ( this.x != targetX ) {
                animationManager.add(new NumberAnimation(this, "x", targetX, duration));
            }
            if ( this.y != targetY ) {
                animationManager.add(new NumberAnimation(this, "y", targetY, duration));
            }
        }
    };

    var Layer = function() {
        this.id = makeId();
        this.elements = [];
        this.visible = true;
    }

    Layer.prototype = {
        add: function(el) {
            this.elements.push(el);
        },

        remove: function(el) {
            for ( var i = 0; i < this.elements.length; i++ ) {
                if ( this.elements[i] == el ) {
                    this.elements.splice(i, 1);
                    break;
                }
            }
        },

        clear: function() {
            this.elements = [];
        },

        render: function(optsIn) {
            if ( !this.visible ) {
                return;
            }
            var opts = {
                parent: this,
                alpha: this.alpha
            };
            if ( optsIn && optsIn.alpha !== undefined ) {
                if ( opts.alpha !== undefined ) {
                    opts.alpha *= optsIn.alpha;
                }
                else {
                    opts.alpha = optsIn.alpha;
                }
            }
            for ( var i = 0; i < this.elements.length; i++ ) {
                this.elements[i].render(opts);
            }
        }
    };    

    var TimeView = function(container) {
        var self,
            el = document.createElement('span');
        el.className = "timeGauge ";
        this.el = el;
        this.totalTime = 1;
        this.currentTime = 0;
        this.maxHeight = container.offsetHeight;
        container.appendChild(el);
    }

    TimeView.prototype = {
        setTotalTime: function(totalTime) {
            this.totalTime = totalTime;
        },

        setRemainingTime: function(remainingTime) {
            this.remainingTime = remainingTime;
            this.el.style.height = Math.floor( ( this.totalTime - this.remainingTime ) * this.maxHeight / this.totalTime ) + "px";
        }
    }

    var MessageView = function(container) {
        var el = document.createElement('span'),
            self;
        el.className = "message ";
        el.draggable = false;
        this.textNode = document.createTextNode("");
        el.appendChild(this.textNode);
        this.el = el;
        container.appendChild(el);

    }

    MessageView.prototype = {
        remove: function() {
            if ( this.el ) {
                this.el.parentNode.removeChild(this.el);
            }
        },

        setText: function(message, type) {
            this.el.className = "message " + type;
            this.el.removeChild(this.textNode);
            this.textNode = document.createTextNode(message);
            this.el.appendChild(this.textNode);
            this.type = type;
        },

        show: function(timeOut) {
            var self = this;
            window.setTimeout(function() {
                self.el.className += " appear";
            }, 0);
            if ( timeOut ) {
                window.setTimeout(this.hide.bind(this), timeOut);
            }
        },

        hide: function() {
            var self = this;
            this.el.className = "message " + this.type + " disappear";
            var realHide = function() {
                self.el.stopObserving("transitionend", realHide);
                self.el.stopObserving("oTransitionEnd", realHide);
                self.el.stopObserving("webkitTransitionEnd", realHide);
                self.el.className += " hidden";
            }
            this.el.observe("transitionend", realHide);
            this.el.observe("oTransitionEnd", realHide);
            this.el.observe("webkitTransitionEnd", realHide);
            //window.setTimeout(this.remove.bind(this), 1000);
        }
    }

    var FieldView = function(container, model, x, y) {
        this.model = model;
        this.id = makeId();
        var spriteType = null;
        if ( model.marbleCount == 1 ) {
            spriteType = sprites.block;
        }
        else if ( model.marbleCount > 1 ) {
            spriteType = sprites.marble;
        }
        
        var el = new Element(spriteType);
        el.x = x * Constants.FIELD_SIZE;
        el.y = y * Constants.FIELD_SIZE;
        container.add(el);

        if ( model.chains ) {
            this.chainsEl = new Element(sprites.chains);
            this.chainsEl.x = el.x;
            this.chainsEl.y = el.y;
        }

        this.el = el;
        this.container = container;
        this._removeClosure = this.remove.bind(this);
    }

    FieldView.prototype = {
        remove: function() {
            this.container.remove(this.el);
            this.el.removed = true;
        },

        destroyAnimation: function() {
            animationManager.add( new NumberAnimation(this.el, "alpha", 0.0, 500) );
            animationManager.add( new NumberAnimation(this.el, "scale", 1.5, 500) );

            setTimeout(this._removeClosure, 1000);
        },

        destroyChain: function() {
            var self = this;
            animationManager.add( new NumberAnimation(this.chainsEl, "alpha", 0.0, 500) );

            setTimeout(function() {
                // TODO: real destroy
                self.chainsEl.visible = false;
                self.chainsEl.removed = true;
            }, 1000);
        }
    }

    var ItemView = function(container, model, x, y, spawning) {
        var self = this;
        this.id = makeId();
        this.container = container;
        this.model = model;
        this.x = x;
        this.y = y;
        this.spawning = spawning;
        var el = new Element(sprites[model.type]);
        this.el = el;
        el.x = x * Constants.FIELD_SIZE;
        el.y = y * Constants.FIELD_SIZE;
        if ( spawning ) {
            el.alpha = 0.0;
            el.scale = 0.0;
            animationManager.add( new NumberAnimation(this.el, "alpha", 1.0, 200) );
            animationManager.add( new NumberAnimation(this.el, "scale", 1.0, 200) );
            setTimeout(function() { self.spawning = false; }, 200 );
        }

        this.id = ItemView.currentId++;
        container.add(el);
        this._removeClosure = this.remove.bind(this);
    }

    ItemView.currentId = 0;
    ItemView.prototype = {
        remove: function() {
            this.container.remove(this.el);
            this.el.removed = true;
        },

        collect: function() {
            var self = this,
                el = this.el;

            animationManager.add(new NumberAnimation(this.el, "scale", 0.0, 300) );
            animationManager.add(new NumberAnimation(this.el, "alpha", 0.0, 300) );
            setTimeout( this._removeClosure, 500 );
        },

        /**
        * Animate movement of this item to the given target board position.
        * @param x target x coordinate
        * @param y target y coordinate
        */
        moveTo: function(x, y) {
            this.el.moveTo(x * Constants.FIELD_SIZE, y * Constants.FIELD_SIZE, 200);
            this.x = x;
            this.y = y;
        },

        failedMoveTo: function(x, y) {
            var self = this,
                targetX = x * Constants.FIELD_SIZE,
                targetY = y * Constants.FIELD_SIZE,
                anims = [],
                bounceFunc = function(x) {
                    var y;
                    if ( x < 0.5 ) {
                        y = x * 2;
                    }
                    else {
                        y = 2.0 - x * 2;
                    }
                    return y;
                };

            if ( this.el.x != targetX ) {
                var anim = new NumberAnimation(this.el, "x", targetX, 400);
                animationManager.add(anim);
                anim.func = bounceFunc;
            }
            if ( this.el.y != targetY ) {
                var anim = new NumberAnimation(this.el, "y", targetY, 400);
                animationManager.add(anim);
                anim.func = bounceFunc;
            }
        }
    }

    var CornerView = function(container, type, x, y) {
        var sprite = cornerSprites[type],
            el = new Element(sprite);
        this.id = makeId();
        el.className = "corner " + type;
        el.x = x * Constants.CORNER_SIZE + sprite.marginX;
        el.y = y * Constants.CORNER_SIZE + sprite.marginY;
        this.el = el;
        container.add(el);
        this.container = container;
        this._removeClosure = this.remove.bind(this);
    }

    CornerView.prototype = {
        remove: function() {
            this.container.remove(this.el);
            this.el.removed = true;
        }
    }

    var BoardView = function(container) {
        this.id = makeId();
        this.el = document.createElement('div');
        this.el.className = "board";
        this.el.style.width = Constants.FIELD_SIZE * Constants.BOARD_WIDTH + "px";
        this.el.style.height = Constants.FIELD_SIZE * Constants.BOARD_HEIGHT + "px";

        this.board = null;
        this.fieldViews = null;
        this.itemViews = null;

        this.background = new Element();

        this.boardLayer = new Layer();
        this.boardLayer.alpha = 0.0;
        this.boardLayer.add(this.background);

        this.fieldsContainer = new Layer();
        this.boardLayer.add(this.fieldsContainer);

        this.itemsContainer = new Layer();
        this.boardLayer.add(this.itemsContainer);

        this.itemsOverlayContainer = new Layer();
        this.boardLayer.add(this.itemsOverlayContainer);

        this.cornersContainer = new Layer();
        this.boardLayer.add(this.cornersContainer);

        container.appendChild(this.el);
    }

    BoardView.prototype = {

        show: function() {
            animationManager.add(new NumberAnimation(this.boardLayer, "alpha", 1.0, 500));
        },

        hide: function() {
            animationManager.add(new NumberAnimation(this.boardLayer, "alpha", 0.0, 500));
        },

        _initFields : function() {
            var fields = this.board.fields;
            var views = [];

            for ( var y = 0; y < fields.length; y++ ) {
                var row = fields[y];
                var viewRow = [];
                for ( var x = 0; x < row.length; x++ ) {
                    var field = row[x];
                    if ( field ) {
                        fieldView = new FieldView(this.fieldsContainer, field, x, y)
                        viewRow.push( fieldView );
                        if ( fieldView.chainsEl ) {
                            // HACK: find a better way
                            this.itemsOverlayContainer.add(fieldView.chainsEl);
                        }
                    }
                    else {
                        viewRow.push(null);
                    }
                }
                views.push(viewRow);
            }
            return views;
        },

        _initItems : function() {
            var fields = this.board.items;
            var views = [];

            for ( var y = 0; y < fields.length; y++ ) {
                var row = fields[y];
                var viewRow = [];
                for ( var x = 0; x < row.length; x++ ) {
                    var field = row[x];
                    if ( field ) {
                        viewRow.push( new ItemView(this.itemsContainer, field, x, y) );
                    }
                    else {
                        viewRow.push(null);
                    }
                }
                views.push(viewRow);
            }
            return views;
        },

        _initCorners : function() {
            var fields = this.board.corners;
            var views = [];

            for ( var y = 0; y < fields.length; y++ ) {
                var row = fields[y];
                var viewRow = [];
                for ( var x = 0; x < row.length; x++ ) {
                    var field = row[x];
                    if ( field ) {
                        viewRow.push( new CornerView(this.cornersContainer, field, x, y) );
                    }
                    else {
                        viewRow.push(null);
                    }
                }
                views.push(viewRow);
            }
            return views;
        },

        _destroyViews : function(views) {
            if ( !views ) {
                return;
            }

            for ( var i = 0; i < views.length; i++ ) {
                for ( var j = 0; j < views[i].length; j++ ) {
                    if ( views[i][j] ) {
                        views[i][j].remove();
                    }
                }
            }
        },

        _destroyFields : function() {
            this.fieldViews = null;
            this.fieldsContainer.clear();
        },

        _destroyItems : function() {
            this.itemViews = null;
            this.itemsContainer.clear();
        },

        _destroyCorners : function() {
            this.cornerViews = null;
            this.cornersContainer.clear();
        },

        _clear: function() {
            if ( this.board ) {
                // TODO: maybe just destroy and recreate the layer?
                this._destroyFields();
                this._destroyItems();
                this._destroyCorners();
                this.itemsOverlayContainer.clear();
            }
        },

        setBoard: function(board) {
            this._clear();
            this.board = board;
            this.fieldViews = this._initFields();
            this.itemViews = this._initItems();
            this.cornerViews = this._initCorners();
            this.setImage(this.board.getImageName());
        },

        getFieldView: function(x, y) {
            return this.fieldViews[y][x];
        },

        setFieldView: function(x, y, view) {
            this.fieldViews[y][x] = view;
        },

        getItemView: function(x, y) {
            return this.itemViews[y][x];
        },

        setItemView: function(x, y, view) {
            this.itemViews[y][x] = view;
        },

        spawnItem: function(x, y, model) {
            var view = new ItemView(this.itemsContainer, model, x, y, true);
            this.setItemView(x, y, view);
        },

        destroyMarble: function(x, y, model) {
            var marbleView = this.getFieldView(x, y);
            if ( marbleView ) {
                marbleView.destroyAnimation();
            }
            var view = new FieldView(this.fieldsContainer, model, x, y);
            this.setFieldView(x, y, view);
        },

        destroyChain: function(x, y) {
            var marbleView = this.getFieldView(x, y);
            if ( marbleView ) {
                marbleView.destroyChain();
            }
        },

        setImage: function(imageName) {
            var self = this;
            renderer.createTexture("img/" + imageName, function(texture) {
                self.background.sprite = renderer.createSprite(texture, 0, 0, 800, 500);
            });
        },

        remove : function() {
            this._clear();
        },

        render: function() {
            this.boardLayer.render();
        }
    }

    var GameView = function(container, mode) {
        var self = this;
        this.container = container;
        this.id = makeId();

        this.boardView = new BoardView(container);

        el = document.createElement('span');
        el.className = "overlay disappear";
        el.draggable = false;
        this.overlayEl = el;
        this.boardView.el.appendChild(el);

        // TODO: use internal element!
        this.scoreEl = document.getElementById("score");

        this.message = new MessageView(this.boardView.el);
        this.timeView = new TimeView(document.getElementsByClassName("timeGaugeContainer")[0]);

        this.dirty = true;
        this._initCanvas(mode);
    }

    GameView.prototype = {
        _initCanvas: function(mode) {
            var self = this,
                canvas = document.createElement("canvas");
            this.boardView.el.appendChild(canvas);
            canvas.width = 800; //canvas.clientWidth;
            canvas.height = 500; //canvas.clientHeight;
            this._nextFrameClosure = this.nextFrame.bind(this);

            if ( !mode || mode == "auto" ) {
                try {
                    console.log("Init WebGLRenderer");
                    renderer = new WebGlRenderer(canvas);
                }
                catch (e) {
                    console.log("Fallback to CanvasRenderer");
                    renderer = new CanvasRenderer(canvas);
                }
            }
            else if ( mode == "canvas" ) {
                console.log("Init CanvasRenderer");
                renderer = new CanvasRenderer(canvas);
            }
            else if ( mode == "webgl" ) {
                console.log("Init WebGlRenderer");
                renderer = new WebGlRenderer(canvas);
            }
            
            initSprites(function() {
                var el = new Element(sprites["cursor"]);
                el.visible = false;
                self.cursorEl = el;
                self.boardView.boardLayer.add(el);

                window.requestAnimFrame(self._nextFrameClosure);
            });
        },

        showBoard: function(visible) {
            this.dirty = true;
            if ( visible ) {
                this.boardView.show();
            }
            else {
                this.boardView.hide();
            }
        },

        setBoard: function(board) {
            this.boardView.setBoard(board);
        },

        setCursor: function(x, y) {
            this.dirty = true;
            this.cursorEl.x = x * Constants.FIELD_SIZE;
            this.cursorEl.y = y * Constants.FIELD_SIZE;
            this.cursorEl.visible = true;
        },

        clearCursor: function() {
            this.dirty = true;
            this.cursorEl.visible = false;
        },

        /**
        * Swap the items at the two given positions.
        */
        swapItems: function(x1, y1, x2, y2) {
            var firstItem = this.boardView.getItemView(x1, y1);
            var secondItem = this.boardView.getItemView(x2, y2);
            this.boardView.setItemView(x1, y1, secondItem);
            this.boardView.setItemView(x2, y2, firstItem);
            firstItem.moveTo(x2, y2);
            secondItem.moveTo(x1, y1);
        },

        /**
        * Try to swap the items at the two given positions,
        * but make it fail (for invalid moves).
        */
        failedSwapItems: function(x1, y1, x2, y2) {
            var firstItem = this.boardView.getItemView(x1, y1);
            var secondItem = this.boardView.getItemView(x2, y2);
            firstItem.failedMoveTo(x2, y2);
            secondItem.failedMoveTo(x1, y1);
        },

        moveItem: function(x1, y1, x2, y2) {
            var item = this.boardView.getItemView(x1, y1);
            if ( !item ) {
                return;
            }
            this.boardView.setItemView(x1, y1, null);
            this.boardView.setItemView(x2, y2, item);
            item.moveTo(x2, y2);
        },

        collectItem: function(x, y) {
            var item = this.boardView.getItemView(x, y);
            item.collect();
            this.boardView.setItemView(x, y, null);
        },

        spawnItem: function(x, y, model) {
            this.boardView.spawnItem(x, y, model);
        },

        destroyMarble: function(x, y, newModel) {
            this.boardView.destroyMarble(x, y, newModel);
        },

        destroyChain: function(x, y) {
            this.boardView.destroyChain(x, y);
        },

        showMessage: function(message, type, timeOut) {
            // TODO: message list ? cached message presets ?
            this.message.setText(message, type || "general");
            this.message.show(timeOut);
        },

        showOverlay: function(show) {
            if ( show ) {
                this.overlayEl.className = "overlay";
            }
            else {
                this.overlayEl.className = "overlay disappear";
            }
        },

        setScore: function(score) {
            this.scoreEl.innerHTML = score;
        },

        setTotalTime: function(totalTime) {
            this.timeView.setTotalTime(totalTime);
        },

        setRemainingTime: function(remainingTime) {
            this.timeView.setRemainingTime(remainingTime);
        },

        render: function() {
            if ( this.dirty ) {
                this.dirty = false;
                renderer.clear();
                this.boardView.render();
            }
        },

        nextFrame: function() {
            if ( animationManager.process()) {
                this.dirty = true;
            };
            this.render();
            window.requestAnimFrame(this._nextFrameClosure);
        }
    }
    
    a3.TimeView = TimeView;
    a3.MessageView = MessageView;
    a3.FieldView = FieldView;
    a3.ItemView = ItemView;
    a3.BoardView = BoardView;
    a3.GameView = GameView;
    
    window.requestAnimFrame = (function(){
        return  window.requestAnimationFrame       ||
                window.webkitRequestAnimationFrame ||
                window.mozRequestAnimationFrame    ||
                window.oRequestAnimationFrame      ||
                window.msRequestAnimationFrame     ||
                function( callback ){
                    window.setTimeout(callback, 1000 / 60);
                };
    })();
})();
