/**
 *  A3
 *
 *  By Vincent Petry - PVince81 at yahoo dot fr
 *
 *  ---------------------------------------------------------------------------
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.*
 */
(function() {

var Field = a3.Field;
var Item = a3.Item;
var Engine = function(gameView) {
    this.board = null;
    this.gameView = gameView;
    this.gameState = new a3.GameState();
    this.levels = new a3.Levels();
}

Engine.prototype = {
    getGameState: function() {
        return this.gameState;
    },

 /**
     * Starts game with the given board.
     */
    start: function() {
        var self = this,
            state = this.gameState;

        
        state.levelNum = 1;
        state.score = 0;
        state.combo = 0;
        state.remainingTime = 0;

        this.gameView.showBoard(false);
        this.gameView.setScore(0);
        this.startLevel();
    },

    startLevel: function() {
        var self = this;
        var board = self.levels.getLevel(self.gameState.levelNum);
        board.populate();
        board.generateCorners();
        this.gameView.setTotalTime(board.getTotalTime());
        this.gameState.remainingTime = board.getTotalTime();
        this.gameView.setRemainingTime(this.gameState.remainingTime);

        this.gameView.showOverlay(true);
        this.gameView.showMessage("Get ready !", null, 1000);
        window.clearInterval(self.secondTimer);
        self.secondTimer = null;

        window.setTimeout(
            function(){
                self.setBoard(board);
                self.gameView.showOverlay(false);
                self.gameView.showBoard(true);
                self.gameState.paused = false;
                self.secondTimer = window.setInterval(self.tickSecond.bind(self), 1000);
                self.gameState.canInteract = true;
                // boards should not have aligned items at start,
                // this is just to be sure
                window.setTimeout(function() {
                    if ( !self.processAligned(self) ) {
                        self.startFall();
                    }
                }, 200);
            }, 1000
        );
    },

    isPaused: function() {
        return !!this.gameState.paused;
    },

    setPaused: function(paused) {
        this.gameState.paused = !!paused;
        this.gameState.canInteract = !paused;
        // TODO: show/hide overlay
    },

    isValidMove: function(x, y, vx, vy) {
        var currentItem = this.board.getItem(x, y),
            // directions to check, excluding the one we're coming from
            directions = [
                [vx, vy],
                [vy, vx],
                [-vy, -vx]
            ],
            // number of identical items in each direction
            count = [0, 0, 0];

        // there must be at least two other items of the same type
        // starting from the new position and following a direction
        for ( var i = 0; i < directions.length; i++ ) {
            var direction = directions[i],
                checkedItem = null,
                // start position
                pos = [x + vx, y + vy];
            // move one
            do {
                count[i]++;
                pos[0] += direction[0];
                pos[1] += direction[1];
                checkedItem = this.board.getItem(pos[0], pos[1]);
            }
            while (checkedItem != Item.TYPE_EMPTY && checkedItem == currentItem);
        }

        // found more than 2 in same direction
        // or in both rotated directions ?
        if ( count[0] >= 3 || count[1] + count[2] >= 4) {
            return true;
        }
        return false;
    },

    processFall: function() {
        var self = this,
            found = false;

        var process = function(board, x, y) {
            // Falling order:
            // 1) Item above
            // 2) Diagonal top-left (only if item not candidate to falling downwards)
            // 3) Diagonal top-right (only if item not candidate to falling downwards)
            // 4) If generator field above, generate new item

            // 1) found an item above ?
            var nextItem = board.getItem(x, y - 1);
            if ( nextItem != Item.TYPE_EMPTY && !board.getField(x, y - 1).chains) {
                // move it down to y
                board.setItem(x, y - 1, null);
                board.setItem(x, y, nextItem);
                // update the view
                self.gameView.moveItem(x, y - 1, x, y);
                return true;
            }
            return false;
        };

        var process2 = function(board, x, y) {

            // check if there's a potentially falling item much higher

            var k = y;
            var item = null;            
            while ( k >= 0 && board.hasMarble(x, k) && (item = board.getItem(x, k)) == Item.TYPE_EMPTY ) {
                k--;
            }

            if ( k >= 0 && item != Item.TYPE_EMPTY && !board.getField(x, k).chains ) {
                // if yes, then go on with the loop
                return false;
            }

            // 2), 3) diagonals, randomly chosen
            var offsets = [
                [-1, -1],
                [1, -1]
            ];
            if ( Math.random() > 0.5 ) {
                offsets.reverse();
            }
            for ( var u = 0; u < offsets.length; u++ ) {
                var xo = x + offsets[u][0];
                var yo = y + offsets[u][1];
                var nextItem = board.getItem(xo, yo);
                if ( nextItem  != Item.TYPE_EMPTY && !board.getField(xo, yo).chains
                    && ( !board.hasMarble(xo, yo + 1) ||
                    board.getItem(xo, yo + 1) != Item.TYPE_EMPTY ) ) {

                    // move it
                    board.setItem(xo, yo, null);
                    board.setItem(x, y, nextItem);
                    // update the view
                    self.gameView.moveItem(xo, yo, x, y);
                    return true;
                }
            }

            // 4) if there's a generator above
            if ( y == 0 || !!board.getField(x, y - 1).generator ) {
                // generate a new item, falling from the top
                nextItem = board.getRandomItem();
                var model = board.setItem(x, y, nextItem);
                self.gameView.spawnItem(x, y, model, nextItem);
                return true;
            }
            return false;
        }

        // from the bottom to the top
        for ( var y = this.board.height - 1; y >= 0; y-- ) {
            for ( var x = 0; x < this.board.width; x++ ) {
                // look for an empty spot
                var item = this.board.getItem(x, y);
                if ( item != Item.TYPE_EMPTY || !this.board.hasMarble(x, y) ) {
                    continue;
                }

                if ( process(this.board, x, y) || process2(this.board, x, y) ) {
                    found = true;
                }
            }
        }

        if (!found) {
            // if none found, cancel interval
            window.clearInterval(this.fallInterval);
            this.fallInterval = null;
            // for combos!
            window.setTimeout(function() {
                self.processAligned();
            }, 0);
            return false;
        }
        return true;
    },

    /**
     * Finds the aligned items and eliminates them.
     */
    processAligned: function() {
        // array of items marked for removal (coordinates)
        var self = this,
            marked = [];

        var getItem0 = function(x, y) {
            return self.board.getItem(x, y);
        };
        var getItem1 = function(y, x) {
            return self.board.getItem(x, y);
        };

        var mark = function(x, y, count, dir) {
            y -= 1;
            do {
                marked.push((dir == 0)?[x, y]:[y, x]);
                y -= 1;
                count--;
            }
            while ( count > 0 );
        };

        var maxCount = 0;
        // loop over 0: "horizontally" and 1: "vertically"
        for ( var dir = 0; dir < 2; dir++ ) {
            var maxX = (dir == 0)?this.board.width:this.board.height,
                maxY = (dir == 0)?this.board.height:this.board.width,
                getItem = (dir == 0)?getItem0:getItem1;

            for ( var x = 0; x < maxX; x++ ) {
                var lastItem = getItem(x, 0),
                    count = 1;

                for ( var y = 1; y < maxY; y++ ) {
                    var item = getItem(x, y);
                    // if same item found, count up
                    // this.hasItem() is to make sure we are not counting falling items
                    if ( lastItem == item && item != Item.TYPE_EMPTY && !(dir == 0 && !this.board.getField(x, y).chains && this.board.hasItem(x, y + 1)) ) {
                        count++;
                    }
                    else {
                        // if we found 3 or more identical items
                        if ( count >= 3 ) {
                            maxCount += count;
                            // mark them for removal, for this we need to go back
                            mark(x, y, count, dir);
                        }
                        // reset counter
                        count = 1;
                        lastItem = item;
                    }
                }
                if ( count >= 3 ) {
                    // mark them for removal, for this we need to go back
                    mark(x, y, count, dir);
                }

            }
        }

        // eliminate marked items
        for ( var i = 0; i < marked.length; i++ ) {
            var x = marked[i][0];
            var y = marked[i][1];
            var item = this.board.getItem(x, y);
            // an item might be removed twice if part of a cross
            if ( item != null ) {
                this.gameView.collectItem(x, y);
                this.board.setItem(x, y, null);

                // if the removed item is on a marble field
                var field = this.board.getField(x, y);
                if ( field && field.chains > 0 ) {
                    // first remove the chains
                    field.chains--;
                    this.gameView.destroyChain(x, y);
                }
                else if ( field && field.marbleCount > 1 ) {
                    var newField = new Field(field);
                    newField.marbleCount--;
                    // and animate
                    this.gameView.destroyMarble(x, y, newField);
                    this.board.setField(x, y, newField);
                }
            }
        }

        if ( maxCount == 3 ) {
            this.addScore(1);
        }
        else if ( maxCount == 4 ) {
            this.addScore(2);
            this.showShortMessage("Four at once !");
        }
        else if ( maxCount == 5 ) {
            this.addScore(5);
            this.showShortMessage("Five at once !");
        }
        else if ( maxCount == 6 ) {
            this.addScore(10);
            this.showShortMessage("Six at once !");
        }
        else if ( maxCount > 6 ) {
            this.addScore(100);
            this.showShortMessage("!!! WOW !!!");
        }

        if ( this.gameState.canInteract && this.board.getMarbleCount() <= 0  ) {
            this.win();
        }

        if ( marked.length > 0 ) {
            // make other items fall down
            this.startFall();
            return true;
        }
        return false;
    },

    startFall: function() {
        var self = this;
        // only start if not started
        if ( !this.fallInterval ) {
            // process fall each second
            this.fallInterval = window.setInterval(function() {
                // FIXME: use a better time-aware way
                self.processFall();
            }, 100);
        }
    },

    addScore: function(inc) {
        this.gameState.score += inc;
        this.gameView.setScore(this.gameState.score);
    },

    tickSecond: function() {
        // FIXME: prevent cheating when using pause mode
        if ( !this.gameState.canInteract || this.gameState.paused ) {
            return;
        }
        this.gameState.remainingTime -= 1;
        if ( this.gameState.remainingTime <= 0 ) {
            this.timeOut();
        }
        this.gameView.setRemainingTime(this.gameState.remainingTime);
    },

    setBoard: function(board) {
        this.board = board;
        this.gameView.setBoard(board);
    },

    /**
     * Try moving the item at the given position according to the given vector.
     * @param x x coordinate of the item
     * @param y y coordinate of the item
     * @param vx x coordinate of the movement vector
     * @param vy y coordinate of the movement vector
     */
    moveItem : function(x, y, vx, vy) {
        // match check for the item to go there or its opposite to come back
        if ( this.isValidMove(x, y, vx, vy) || this.isValidMove(x + vx, y + vy, -vx, -vy) ) {
            // swap items
            this.board.swapItems(x, y, x + vx, y + vy);
            this.gameView.swapItems(x, y, x + vx, y + vy);

            this.processAligned();
        }
        else {
            // just shake the items
            this.gameView.failedSwapItems(x, y, x + vx, y + vy);
        }
    },


    showMessage: function(text) {
        this.gameView.showMessage(text);
    },

    showShortMessage: function(text) {
        // FIXME: this message doesn't appear at the right position
        //this.gameView.showMessage(text, "short", 500);
    },

    timeOut: function() {
        var self = this;
        window.clearInterval(self.secondTimer);
        self.secondTimer = null;
        this.gameState.canInteract = false;
        self.gameView.showOverlay(true);
        self.gameView.showBoard(false);
        self.showMessage("Time out !", null, 2000);
        window.setTimeout(self.startLevel.bind(self), 3000);
    },

    win: function() {
        var self = this;
        window.clearInterval(self.secondTimer);
        self.secondTimer = null;

        this.gameState.canInteract = false;
        window.setTimeout(function doWin() {
            self.gameView.showOverlay(true);
            self.gameView.showBoard(false);
            self.showMessage("You win!", null, 2000);
            self.gameState.levelNum++;
            window.setTimeout(self.startLevel.bind(self), 3000);
        }, 1000 );
    },

    canInteract: function() {
        return this.gameState.canInteract;
    }
}

a3.Engine = Engine;
})();