/**
 *  A3
 *
 *  By Vincent Petry - PVince81 at yahoo dot fr
 *
 *  ---------------------------------------------------------------------------
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.*
 */

(function(){

/**
 * Field model.
 * Attributes are:
 * - marbleCount: 2 if it has a breakable marble, 1 if has a block, 0 otherwise
 * - chains: 1 if has chains, 0 otherwise
 * - generator: true if field is a generator (invisible)
 */
var Field = function(attrs) {
    if ( !attrs ) {
        attrs = {};
    }
    this.marbleCount = attrs.marbleCount || 0;
    this.generator = !!attrs.generator;
    this.chains = attrs.chains || 0;
}

Field.prototype = {
        isEmpty: function() {
            return !this.marbleCount && !this.generator;
        }
}

var EMPTY_FIELD = new Field();

var Item = function(type) {
    this.type = type;
    this.movement = null;
}

Item.prototype = {
}

Item.TYPE_EMPTY = null;
Item.TYPE_NOPOPULATE = "nopopulate"; // meta, to prevent populating at start
Item.TYPE_STAR = "star";
Item.TYPE_TRIANGLE = "triangle";
Item.TYPE_RECTANGLE = "rectangle";
Item.TYPE_CIRCLE = "circle";
Item.TYPE_HEART = "heart";
Item.TYPES = [
    Item.TYPE_STAR,
    Item.TYPE_TRIANGLE,
    Item.TYPE_RECTANGLE,
    Item.TYPE_CIRCLE,
    Item.TYPE_HEART
];

var Corner = {};
Corner.TYPE_EMPTY = null;
Corner.TYPE_TOPLEFT = "topleft";
Corner.TYPE_TOPRIGHT = "topright";
Corner.TYPE_BOTTOMLEFT = "bottomleft";
Corner.TYPE_BOTTOMRIGHT = "bottomright";
Corner.TYPE_TOP = "top";
Corner.TYPE_BOTTOM = "bottom";
Corner.TYPE_LEFT = "left";
Corner.TYPE_RIGHT = "right";
Corner.TYPE_OUTER_TOPLEFT = "outertopleft";
Corner.TYPE_OUTER_TOPRIGHT = "outertopright";
Corner.TYPE_OUTER_BOTTOMLEFT = "outerbottomleft";
Corner.TYPE_OUTER_BOTTOMRIGHT = "outerbottomright";
Corner.TYPE_CROSS = "cross";
Corner.TYPE_CROSS2 = "cross2";

var Animation = function(targetObject) {
    this.targetObject = targetObject;
}

Animation.prototype = {
    update: function(step) {
        // TODO
    }
}

var Board = function() {
    function makeMatrix(width, height, item) {
        var rows = [];
        for ( var y = 0; y < height; y++ ) {
            var row = [];
            for ( var x = 0; x < width; x++ ) {
                row.push(item);
            }
            rows.push(row);
        }
        return rows;
    }

    this.width = a3.Constants.BOARD_WIDTH;
    this.height = a3.Constants.BOARD_HEIGHT;
    this.fields = makeMatrix(this.width, this.height, EMPTY_FIELD);
    this.items = makeMatrix(this.width, this.height, Item.TYPE_EMPTY);
    this.corners = makeMatrix(this.width * 2 + 2, this.height * 2 + 2, Corner.TYPE_EMPTY);
    this.imageName = null;
    this.marbleCount = 0;
    // total time in seconds
    this.totalTime = 60;
}

Board.prototype = {
    inBounds: function(x, y) {
        var good = ( x >= 0 && y >= 0 && x < this.width && y < this.height );
        /*
        if ( !good ) {
            console.warn("Out of bounds: ", x, y);
        }
        */
        return good;
    },

    getMarbleCount: function() {
        return this.marbleCount;
    },
    
    setField: function(x, y, newField, oldField) {
        if ( !this.inBounds(x,y) ) {
            return null;
        }

        var oldField = oldField || this.getField(x, y);
        if ( oldField && oldField.marbleCount > 1 ) {
            oldField.marbleCount--;
            this.marbleCount--;
        }

        if ( !newField || newField.isEmpty() ) {
            return this.fields[y][x] = null;
        }
        else {
            if ( newField.marbleCount > 1 ) {
                this.marbleCount+= newField.marbleCount - 1;
            }
            return this.fields[y][x] = newField;
        }
    },

    getField: function(x, y) {
        if ( !this.inBounds(x,y) ) {
            return null;
        }

        var field = this.fields[y][x];
        if ( !field ) {
            return EMPTY_FIELD;
        }
        return field;
    },

    setItem: function(x, y, type) {
        if ( !this.inBounds(x,y) ) {
            return;
        }

        if ( !type || type == Item.TYPE_EMPTY ) {
            return this.items[y][x] = null;
        }
        else {
            return this.items[y][x] = new Item(type);
        }
    },

    getItem: function(x, y) {
        if ( !this.inBounds(x,y) ) {
            return null;
        }

        var item = this.items[y][x];
        if ( !item ) {
            return Item.TYPE_EMPTY;
        }
        return item.type;
    },

    swapItems: function(x1, y1, x2, y2) {
        var firstItem = this.getItem(x1, y1);
        var secondItem = this.getItem(x2, y2);
        this.setItem(x1, y1, secondItem);
        this.setItem(x2, y2, firstItem);
    },

    getRandomItem: function(exclusions) {
        var index = parseInt(Math.floor(Math.random()*Item.TYPES.length));
        if ( !exclusions || exclusions.length == 0 ) {
            return Item.TYPES[index];
        }
        else {
            // as long as it's an excluded item
            while (exclusions.indexOf(Item.TYPES[index]) >= 0) {
                index++;
                if ( index >= Item.TYPES.length ) {
                    // loop
                    index = 0;
                }
            }
            return Item.TYPES[index];
        }
    },

    /**
     * Populates the empty fields with items.
     */
    populate: function() {
        for ( var y = 0; y < this.height; y++ )
        for ( var x = 0; x < this.width; x++ ) {
            if ( this.hasMarble(x, y) && this.getItem(x, y) == Item.TYPE_EMPTY )  {
                // check two items left and two items up
                var itemUp1 = this.getItem( x, y - 1 );
                var itemUp2 = this.getItem( x, y - 2 );
                var itemLeft1 = this.getItem( x - 1, y );
                var itemLeft2 = this.getItem( x - 2, y );
                var exclusions = [];
                if ( itemUp1 == itemUp2 ) {
                    exclusions.push(itemUp1);
                }
                if ( itemLeft1 == itemLeft2 ) {
                    exclusions.push(itemLeft1);
                }

                // generate item
                var itemType = this.getRandomItem(exclusions);
                if ( !itemType  || itemType == Item.EMPTY ) {
                    console.warn("Populated with an empty item !");
                }
                this.setItem(x, y, itemType);
            }
            if ( this.getItem(x, y) == Item.TYPE_NOPOPULATE ) {
                this.setItem(x, y, null);
            }
        }
    },

    generateCorners: function() {
        for ( var x = 0; x < this.width + 1; x++ )
        for ( var y = 0; y < this.height + 1; y++ ) {
            var tx = x * 2;
            var ty = y * 2;
            // TODO: simply / factorize this logic
            if ( this.hasMarble( x, y ) )  {
                var hasTop = !this.hasMarble( x, y - 1 );
                var hasLeft = !this.hasMarble( x - 1, y );
                var hasTopLeft = !this.hasMarble( x - 1, y - 1 );
                if ( hasTop ) {
                    this.corners[ty][tx] = Corner.TYPE_TOP;
                    this.corners[ty][tx + 1] = Corner.TYPE_TOP;
                }
                if ( hasLeft ) {
                    this.corners[ty][tx] = hasTop?Corner.TYPE_TOPLEFT:Corner.TYPE_LEFT;
                    this.corners[ty + 1][tx] = Corner.TYPE_LEFT;
                }
                if ( hasTopLeft && !hasLeft && !hasTop ) {
                    this.corners[ty][tx] = Corner.TYPE_OUTER_BOTTOMRIGHT;
                }

                if ( !hasTopLeft ) {
                    if ( !hasTop && hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_OUTER_TOPRIGHT;
                    }
                    if ( hasTop && !hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_OUTER_BOTTOMLEFT;
                    }
                    if ( hasTop && hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_CROSS;
                    }
                }
            }
            else {
                var hasTop = this.hasMarble( x, y - 1 );
                var hasLeft = this.hasMarble( x - 1, y );
                var hasTopLeft = this.hasMarble( x - 1, y - 1 );
                if ( hasTop ) {
                    this.corners[ty][tx] = Corner.TYPE_BOTTOM;
                    this.corners[ty][tx + 1] = Corner.TYPE_BOTTOM;
                }
                if ( hasLeft ) {
                    this.corners[ty][tx] = hasTop?Corner.TYPE_BOTTOMRIGHT:Corner.TYPE_RIGHT;
                    this.corners[ty + 1][tx] = Corner.TYPE_RIGHT;
                }
                if ( hasTopLeft ) {
                    if ( !hasTop && !hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_BOTTOMRIGHT;
                    }
                    if ( hasTop && hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_OUTER_TOPLEFT;
                    }
                }
                if ( !hasTopLeft ) {
                    if ( hasTop ) {
                        this.corners[ty][tx] = Corner.TYPE_BOTTOMLEFT;
                    }
                    if ( hasLeft ) {
                        this.corners[ty][tx] = Corner.TYPE_TOPRIGHT;
                    }
                    if ( hasLeft && hasTop) {
                        this.corners[ty][tx] = Corner.TYPE_CROSS2;
                    }
                }
            }
        }
    },
 
    setImageName: function(imageName) {
        this.imageName = imageName;
    },

    getImageName: function() {
        return this.imageName;
    },

    setTotalTime: function(totalTime) {
        this.totalTime = totalTime;
    },

    getTotalTime: function() {
        return this.totalTime;
    },

    /**
     * Returns whether the item at the given position has
     * an empty space below.
     */
    hasItem: function(x, y) {
        var field = this.getField(x, y);
        return (field && field.marbleCount > 0 && this.getItem(x,y) == Item.TYPE_EMPTY);
    },

    hasMarble: function(x, y) {
        var field = this.getField(x, y);
        return field && field.marbleCount > 0;
    },

    canMoveItem: function(x, y) {
        return this.getItem(x, y) != Item.TYPE_EMPTY && !this.getField(x, y).chains;
    }
}

var GameState = function() {
    this.score = 0;
    this.canInteract = false;
    this.paused = true;
    this.levelNum = 1;
    this.combo = 0;
    this.remainingTime = 0;
}

GameState.prototype = {
}

a3.Field = Field;
a3.Item = Item;
a3.Animation = Animation;
a3.Board = Board;
a3.GameState = GameState;

})();