/**
 *  Web GL sprites library
 *
 *  By Vincent Petry - PVince81 at yahoo dot fr
 *
 *  ---------------------------------------------------------------------------
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.*
 */
(function() {
    FRAGMENT_SHADER =
        "#ifdef GL_ES\n" +
        "precision highp float;\n" +
        "#endif\n" +

        "uniform sampler2D uSampler;" +
        "varying vec2 vTextureCoord;" +
        "uniform float uAlpha;" +

        "void main(void)" +
        "{" +
        "   vec4 textureColor = texture2D(uSampler, vec2(vTextureCoord.s, vTextureCoord.t));" +
        "   gl_FragColor = vec4(textureColor.rgb, textureColor.a * uAlpha);" +
        "}";
    VERTEX_SHADER =
        "attribute vec3 aVertexPosition;" +
        "attribute vec2 aTextureCoord;" +

        "uniform mat4 uMVMatrix;" +
        "uniform mat4 uPMatrix;" +

        "varying vec2 vTextureCoord;" +

        "void main(void)" +
        "{" +
        "    gl_Position = uPMatrix * uMVMatrix * vec4(aVertexPosition, 1.0);" +
        "    vTextureCoord = aTextureCoord;" +
        "}";

    function createContext(canvas) {
        var gl = null;
        var errors = "";
        try {
            gl = canvas.getContext("webgl");
            if ( gl != null ) {
               return gl;
            }
        }
        catch ( e ) {
        }

        try {
            gl = canvas.getContext("experimental-webgl");
        }
        catch ( e ) {
        }

        // test for older versions
        try {
            var test = new Float32Array([]);
        }
        catch (e) {
            Float32Array = WebGLFloatArray;
        }

        return gl;
    };

    var Renderer = function(canvas) {
        var gl = createContext(canvas);
        if ( !gl ) {
            throw "WebGL init failed";
        }

        this._mvMatrix = mat4.create();
        this._pMatrix = mat4.create();
        mat4.identity(this._mvMatrix);
        mat4.identity(this._pMatrix);

        gl.viewportWidth = canvas.clientWidth;
        gl.viewportHeight = canvas.clientHeight;
        gl.viewport(0, 0, gl.viewportWidth, gl.viewportHeight);
        mat4.ortho(0, gl.viewportWidth, gl.viewportHeight, 0, 0, 1, this._pMatrix);

        gl.clearColor(0, 0, 0, 1.0);
        gl.disable(gl.DEPTH_TEST);
        gl.enable(gl.BLEND);
        gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
        this.gl = gl;
        this._initShaders();
        this._initBuffers();

        // initializing these here as they won't change
        gl.bindBuffer(gl.ARRAY_BUFFER, this._blockVertexBuf);
        gl.vertexAttribPointer(
            this._shaderProgram.vertexPositionAttribute,
            this._blockVertexBuf.itemSize, gl.FLOAT, false, 0, 0);

        gl.activeTexture(gl.TEXTURE0);
        gl.uniform1i(this._shaderProgram.samplerUniform, 0);
        gl.uniformMatrix4fv(this._shaderProgram.pMatrixUniform, false, this._pMatrix);
    };

    Renderer.prototype = {

        _initBuffers : function() {
            var gl = this.gl;
            this._blockVertexBuf = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this._blockVertexBuf);
            vertices = [
            // Single face
            0.0, 0.0, 0.0,
            1.0, 0.0, 0.0,
            0.0, 1.0, 0.0,
            1.0, 1.0, 0.0
            ];

            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertices), gl.STATIC_DRAW);

            this._blockVertexBuf.itemSize = 3;
            this._blockVertexBuf.numItems = vertices.length / 3;

            this.texCoords = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this.texCoords);
            var texCoords = [
                0.0, 0.0,
                1.0, 0.0,
                0.0, 1.0,
                1.0, 1.0
            ];

            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(texCoords),
                    gl.STATIC_DRAW);
            this.texCoords.itemSize = 2;
            this.texCoords.numItems = texCoords.length;
        },

        _getShader : function(gl, type, script) {
            var shader = gl.createShader(type);
            gl.shaderSource(shader, script);
            gl.compileShader(shader);

            if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
                console.error("Shader compile error: ", gl.getShaderInfoLog(shader));
                throw "Shader compile error: " + gl.getShaderInfoLog(shader);
            }

            return shader;
        },

        _initShaders : function() {
            var gl = this.gl;

            var fragmentShader = this._getShader(gl, gl.FRAGMENT_SHADER, FRAGMENT_SHADER);
            var vertexShader = this._getShader(gl, gl.VERTEX_SHADER, VERTEX_SHADER);

            this._shaderProgram = gl.createProgram();
            gl.attachShader(this._shaderProgram, vertexShader);
            gl.attachShader(this._shaderProgram, fragmentShader);
            gl.linkProgram(this._shaderProgram);

            if (!gl.getProgramParameter(this._shaderProgram, gl.LINK_STATUS)) {
                throw "Could not initialise shaders";
            }

            gl.useProgram(this._shaderProgram);

            this._shaderProgram.vertexPositionAttribute = gl.getAttribLocation(this._shaderProgram, "aVertexPosition");
            gl.enableVertexAttribArray(this._shaderProgram.vertexPositionAttribute);

            this._shaderProgram.pMatrixUniform = gl.getUniformLocation(this._shaderProgram, "uPMatrix");
            this._shaderProgram.mvMatrixUniform = gl.getUniformLocation(this._shaderProgram, "uMVMatrix");
            this._shaderProgram.samplerUniform = gl.getUniformLocation(this._shaderProgram, "uSampler");
            this._shaderProgram.aTextureCoord = gl.getAttribLocation(this._shaderProgram, "aTextureCoord");
            this._shaderProgram.uAlpha = gl.getUniformLocation(this._shaderProgram, "uAlpha");
            gl.enableVertexAttribArray(this._shaderProgram.aTextureCoord);
        },

        drawTexture : function(texture, x, y, width, height) {
            var gl = this.gl;
            mat4.identity(this._mvMatrix);
            mat4.translate(this._mvMatrix, [x, y, 0.0]);
            mat4.scale(this._mvMatrix,[width,height,0.0]);

            gl.bindBuffer(gl.ARRAY_BUFFER, this.texCoords);
            gl.vertexAttribPointer(
                this._shaderProgram.aTextureCoord,
                this.texCoords.itemSize, gl.FLOAT, false, 0, 0);
            gl.bindTexture(gl.TEXTURE_2D, texture._texture);

            gl.uniformMatrix4fv(this._shaderProgram.mvMatrixUniform, false, this._mvMatrix);
            gl.drawArrays(gl.TRIANGLE_STRIP, 0, this._blockVertexBuf.numItems);
        },

        drawSprite : function(sprite, x, y, alpha, scale) {
            var gl = this.gl;
            mat4.identity(this._mvMatrix);
            mat4.translate(this._mvMatrix, [x, y, 0.0]);
            if ( scale === undefined || scale === 1.0 ) {
                scale = 1.0;
                mat4.scale(this._mvMatrix,[sprite.width,sprite.height,0.0]);
            }
            else {
                var newWidth = sprite.width * scale,
                    newHeight = sprite.height * scale;
                
                mat4.translate(this._mvMatrix, [(sprite.width - newWidth) / 2, (sprite.height - newHeight) / 2, 0.0]);
                mat4.scale(this._mvMatrix,[newWidth,newHeight,0.0]);
            }

            gl.bindBuffer(gl.ARRAY_BUFFER, sprite.texCoords);
            gl.vertexAttribPointer(
                this._shaderProgram.aTextureCoord,
                this.texCoords.itemSize, gl.FLOAT, false, 0, 0);
            gl.bindTexture(gl.TEXTURE_2D, sprite.texture._texture);

            if ( alpha === undefined ) {
                alpha = 1.0;
            }
            
            gl.uniform1f(this._shaderProgram.uAlpha, alpha);

            gl.uniformMatrix4fv(this._shaderProgram.mvMatrixUniform, false, this._mvMatrix);
            gl.drawArrays(gl.TRIANGLE_STRIP, 0, this._blockVertexBuf.numItems);            
        },

        createTexture: function(imageSrc, callback) {
            return new Texture(this.gl, imageSrc, callback);
        },

        createSprite: function(texture, x, y, width, height) {
            return new Sprite(texture, x, y, width, height);
        },

        clear: function() {
            this.gl.clear(this.gl.COLOR_BUFFER_BIT);
        }
    };

    var Texture = function(gl, imageSrc, callback) {
        var self = this;
        this.gl = gl;
        this._texture = gl.createTexture();
        this.loaded = false;
        this._image = new Image();
        this._image.onload = function() {
            self._handleLoadedTexture();
            if ( callback ) {
                callback(self);
            }
        }
        this._image.src = imageSrc;
    };

    Texture.prototype = {
        _handleLoadedTexture: function() {
            var gl = this.gl;
            // note: image dimensions must always be a power of two!
            gl.bindTexture(gl.TEXTURE_2D, this._texture);
            gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, this._image);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
            gl.bindTexture(gl.TEXTURE_2D, null);
            this.loaded = true;
        }
    };

    var Sprite = function(texture, x, y, width, height) {
        this.texture = texture;
        this.x = x;
        this.y = y;
        this.width = width;
        this.height = height;
        this._initTexCoords();
    }

    Sprite.prototype = {
        _initTexCoords: function() {
            var gl = this.texture.gl,
                texWidth = this.texture._image.width,
                texHeight = this.texture._image.height,
                x1 = this.x / texWidth,
                y1 = this.y / texHeight,
                x2 = (this.x + this.width ) / texWidth,
                y2 = (this.y + this.height ) / texHeight;

            this.texCoords = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this.texCoords);
            var texCoords = [
                x1, y1,
                x2, y1,
                x1, y2,
                x2, y2
            ];

            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(texCoords),
                    gl.STATIC_DRAW);
            this.texCoords.itemSize = 2;
            this.texCoords.numItems = texCoords.length;
        }
    };

    window.WebGlRenderer = Renderer;

})();